__all__ = [ "pdrutils", "measurement", "modelset"]

VERSION = "2.0.1rc3"
AUTHORS =  'Marc W. Pound, Mark G. Wolfire'
DESCRIPTION="PhotoDissociation Region Toolbox (PDRT), astrophysics analysis tools"
KEYWORDS = "PDR photodissociation astronomy astrophysics"

def version():
    """Version of the PDRT code

    :rtype: str
    """
    return VERSION


